import asyncio
import hashlib
import hmac
import json
import os
import random
import time
from urllib.parse import urljoin

import aiolimiter
import async_timeout
import httpx
import websockets

RATE_LIMITS = {
    # order methods
    (
        "private/create-order",
        "private/cancel-order",
        "private/cancel-all-orders",
        "private/margin/create-order",
        "private/margin/cancel-order",
        "private/margin/cancel-all-orders",
    ): (14, 0.1),
    # order detail methods
    (
        "private/get-order-detail",
        "private/margin/get-order-detail",
    ): (29, 0.1),
    # general trade methods
    (
        "private/get-trades",
        "private/margin/get-trades",
        "private/get-order-history",
        "private/margin/get-order-history",
    ): (1, 1),
}


class ApiError(Exception):
    pass


class ApiListenAsyncIterable:
    def __init__(self, api, ws, channels, sign):
        self.api = api
        self.ws = ws
        self.channels = channels

        self.sign = sign
        self.sub_data_sent = False
        self.auth_sent = False

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.sub_data_sent:
            sub_data = {
                "id": random.randint(1000, 10000),
                "method": "subscribe",
                "params": {"channels": self.channels},
                "nonce": int(time.time()),
            }

        # [0] if not sign start connection with subscription
        if not self.sub_data_sent and not self.sign:
            await self.ws.send(json.dumps(sub_data))
            self.sub_data_sent = True

        async with async_timeout.timeout(60) as tm:
            data = await self.ws.recv()
            tm.shift(60)

        if data:
            data = json.loads(data)
            result = data.get("result")

            # [1] send heartbeat to keep connection alive
            if data["method"] == "public/heartbeat":
                await self.ws.send(
                    json.dumps(
                        {
                            "id": data["id"],
                            "method": "public/respond-heartbeat",
                        }
                    )
                )
            elif self.sub_data_sent and result:
                # [4] consume data
                return result

            # [2] sign auth request to listen private methods
            if self.sign and not self.auth_sent:
                await self.ws.send(
                    json.dumps(self.api.sign("public/auth", {}))
                )
                self.auth_sent = True

            # [3] subscribe to channels
            if (
                data["method"] == "public/auth"
                and data["code"] == 0
                and not self.sub_data_sent
            ):
                await self.ws.send(json.dumps(sub_data))
                self.sub_data_sent = True


class ApiProvider:
    """Provides HTTP-api requests and websocket requests."""

    def __init__(
        self,
        *,
        api_key="",
        api_secret="",
        from_env=False,
        auth_required=True,
        timeout=5,
        retries=6,
        root_url="https://api.crypto.com/v2/",
        ws_root_url="wss://stream.crypto.com/v2/",
        logger=None,
    ):
        self.ssl_context = httpx.create_ssl_context()
        self.api_key = api_key
        self.api_secret = api_secret
        self.root_url = root_url
        self.ws_root_url = ws_root_url
        self.timeout = timeout
        self.retries = retries
        self.last_request_path = ""

        self.rate_limiters = {}

        for urls in RATE_LIMITS:
            for url in urls:
                self.rate_limiters[url] = aiolimiter.AsyncLimiter(
                    *RATE_LIMITS[urls]
                )

        # limits for not matched methods
        self.general_private_limit = aiolimiter.AsyncLimiter(3, 0.1)
        self.general_public_limit = aiolimiter.AsyncLimiter(100, 1)

        if not auth_required:
            return

        if from_env:
            self.read_keys_from_env()

        if not self.api_key or not self.api_secret:
            raise ValueError("Provide api_key and api_secret")

    def read_keys_from_env(self):
        self.api_key = os.environ.get("CRYPTOCOM_API_KEY", "")
        if not self.api_key:
            raise ValueError("Provide CRYPTOCOM_API_KEY env value")
        self.api_secret = os.environ.get("CRYPTOCOM_API_SECRET", "")
        if not self.api_secret:
            raise ValueError("Provide CRYPTOCOM_API_SECRET env value")

    def sign(self, path, data):
        data = data or {}
        data["method"] = path

        sign_time = int(time.time() * 1000)
        data.update({"nonce": sign_time, "api_key": self.api_key})

        data["id"] = random.randint(1000, 10000)
        data_params = data.get("params", {})
        params = "".join(
            f"{key}{data_params[key]}" for key in sorted(data_params)
        )

        payload = f"{path}{data['id']}{self.api_key}{params}{data['nonce']}"

        data["sig"] = hmac.new(
            self.api_secret.encode("utf-8"),
            payload.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()
        return data

    def get_limiter(self, path):
        if path in self.rate_limiters.keys():
            return self.rate_limiters[path]
        else:
            if path.startswith("private"):
                return self.general_private_limit
            elif path.startswith("public"):
                return self.general_public_limit
            else:
                raise ApiError(f"Wrong path: {path}")

    async def request(self, method, path, params=None, data=None, sign=False):
        original_data = data
        limiter = self.get_limiter(path)

        for count in range(self.retries):
            client = httpx.AsyncClient(
                timeout=httpx.Timeout(timeout=self.timeout),
                verify=self.ssl_context,
            )
            if sign:
                data = self.sign(path, original_data)
            try:
                async with limiter:
                    resp = await client.request(
                        method,
                        urljoin(self.root_url, path),
                        params=params,
                        json=data,
                        headers={"content-type": "application/json"},
                    )
                    resp_json = resp.json()
                    if resp.status_code != 200:
                        raise ApiError(
                            f"Error: {resp_json}. "
                            f"Status: {resp.status_code}. Json params: {data}"
                        )
            except (
                asyncio.TimeoutError,
                httpx.HTTPError,
                json.JSONDecodeError,
            ) as exc:
                if count == self.retries:
                    raise ApiError(
                        f"Timeout or read error, retries: {self.retries}. "
                        f"Path: {path}. Data: {data}. Exc: {exc}"
                    ) from exc
                continue
            finally:
                await client.aclose()

            if resp_json["code"] == 0:
                result = resp_json.get("result", {})
                return result.get("data", {}) or result

            if count == self.retries:
                raise ApiError(
                    f"System error, retries: {self.retries}. "
                    f"Code: {resp.status_code}. Json: {resp_json}. "
                    f"Data: {data}"
                )

    async def get(self, path, params=None, sign=False):
        return await self.request("get", path, params=params, sign=sign)

    async def post(self, path, data=None, sign=True):
        return await self.request("post", path, data=data, sign=sign)

    async def listen(self, url, *channels, sign=False):
        url = urljoin(self.ws_root_url, url)
        async for ws in websockets.connect(url, open_timeout=self.timeout):
            try:
                dataiterator = ApiListenAsyncIterable(self, ws, channels, sign)
                async for data in dataiterator:
                    if data:
                        yield data
            except (websockets.ConnectionClosed, asyncio.TimeoutError):
                continue



def round_up(value: float, precision: int) -> float:
    """Rounds number to upper precision 0.13 -> 0.2"""
    pwr = 10**precision
    rounded = round((value * pwr) / pwr, precision)
    if rounded < value:
        rounded = round((value * pwr + 1) / pwr, precision)
    return rounded


def round_down(value: float, precision: int) -> float:
    """Rounds number to lower precision 0.13 -> 0.1"""
    pwr = 10**precision
    return round((value * pwr) / pwr, precision)



import asyncio
import platform

from . import coins, pairs
from .api import ApiError, ApiProvider
from .market import Exchange
from .private import Account
from .structs import (
    Candle,
    Coin,
    Deposit,
    DepositStatus,
    MarketTrade,
    Order,
    OrderSide,
    OrderStatus,
    OrderType,
    Pair,
    Period,
    PrivateTrade,
    Timeframe,
    Withdrawal,
    WithdrawalStatus,
)

if platform.system() == "Windows":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

__all__ = [
    "Order",
    "OrderSide",
    "OrderStatus",
    "OrderType",
    "pairs",
    "Pair",
    "coins",
    "Coin",
    "Period",
    "Candle",
    "MarketTrade",
    "PrivateTrade",
    "Timeframe",
    "Deposit",
    "Withdrawal",
    "DepositStatus",
    "WithdrawalStatus",
    "Exchange",
    "Account",
    "ApiError",
    "ApiProvider",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cryptocom', 'cryptocom.exchange']

package_data = \
{'': ['*']}

install_requires = \
['aiolimiter>=1.0.0,<2.0.0',
 'async-timeout>=4.0.2,<5.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'httpx>=0.22.0,<0.23.0',
 'websockets>=10.3,<11.0']

setup_kwargs = {
    'name': 'cryptocom-exchange',
    'version': '0.11.5',
    'description': 'Python 3.7+ async library for crypto.com/exchange API using httpx and websockets',
    'long_description': "# Python 3.7+ async library for crypto.com/exchange API using httpx and websockets\n\n[![Docs Build Status](https://readthedocs.org/projects/cryptocom-exchange/badge/?version=latest&style=flat)](https://readthedocs.org/projects/cryptocom-exchange)\n![Test workflow](https://github.com/goincrypto/cryptocom-exchange/actions/workflows/test.yml/badge.svg)\n[![Maintainability](https://api.codeclimate.com/v1/badges/8d7ffdae54f3c6e86b5a/maintainability)](https://codeclimate.com/github/goincrypto/cryptocom-exchange/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/8d7ffdae54f3c6e86b5a/test_coverage)](https://codeclimate.com/github/goincrypto/cryptocom-exchange/test_coverage)\n[![PyPI implementation](https://img.shields.io/pypi/implementation/cryptocom-exchange.svg)](https://pypi.python.org/pypi/cryptocom-exchange/)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/cryptocom-exchange.svg)](https://pypi.python.org/pypi/cryptocom-exchange/)\n[![PyPI license](https://img.shields.io/pypi/l/cryptocom-exchange.svg)](https://pypi.python.org/pypi/cryptocom-exchange/)\n[![PyPI version fury.io](https://badge.fury.io/py/cryptocom-exchange.svg)](https://pypi.python.org/pypi/cryptocom-exchange/)\n[![PyPI download month](https://img.shields.io/pypi/dm/cryptocom-exchange.svg)](https://pypi.python.org/pypi/cryptocom-exchange/)\n[![Gitter](https://badges.gitter.im/goincrypto/cryptocom-exchange.svg)](https://gitter.im/goincrypto/cryptocom-exchange?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge)\n\nDocumentation: [https://cryptocom-exchange.rtfd.io](https://cryptocom-exchange.rtfd.io)\n\nExchange original API docs: [https://exchange-docs.crypto.com](https://exchange-docs.crypto.com)\n\n### **WARNING: PLEASE use 0.10.3+ [prev versions have private websocket broken and memory leaks]\n\n### Description\n\n`pip install cryptocom-exchange`\n\n- provides all methods to access crypto.com/exchange API (except for websockets temporary)\n- full test coverage on real exchange with real money\n- simple async methods with custom retries and timeouts\n\n**Please do not use secret keys, they used only for test purposes**\n\n### Changelog\n\n- **0.10.4** - Updated pairs\n- **0.10.3** - fixed websocket private endpoints not returning data\n- **0.10.2** - fixed huge memory leak by `httpx`\n- [leaks memory] **0.10.1** - added read timeouts for websockets, fixed test with tickers\n- [leaks memory] **0.10.0** - moved into httpx + websockets, added autoreconnect, simplified code, improved stability\n- **0.9.5** - added timeout for websocket if no data received in 3 mins we trying to reconnect\n- **0.9.4** - fixed spread func, fixed missing params\n- **0.9.3** - added RPS limiter by @Irishery\n- **0.9.2** - fixed event loop import level\n- **0.9.1** - fixed Windows bug with asyncio event loop\n- **0.9.0** - updated coins, refactored wallet transactions\n- **0.8.1** - fixed coin name generation\n- **0.8** - fixed tests with updated coins\n- **0.7.12** - updated coins, added websocket timeouts\n- **0.7.11** - fixed orders history if empty\n- **0.7.10** - updated pairs precision\n- **0.7.9** - fixed price and quantity rounding, updated pairs and coins\n- **0.7.8** - changed keys, removed depth from orderbook (not working, always 150), updated pairs\n- **0.7.7** - updated pairs by JOB, fixed timeout\n- **0.7.6** - updated pairs\n- **0.7.5** - fixed `order.remaining_quantity` rounding\n- **0.7.4** - fixed sync pairs for account\n- **0.7.3** - fixed price of order if not filled, updated coins, added missing trades to `Order`\n- **0.7.2** - fixed `listen_orders` private account method, added test\n- **0.7.1** - fixed missing '.0' in order price and quantity (different in py3.7, py3.9)\n- **0.7** - major changes, `Pair` -> `cro.pairs.CRO_USDT` moved to more complex structure so we can use round and server information about pairs.\n    - If you have errors, just use `await account.sync_pairs()` or `await exchange.sync_pairs()`\n    - Added rounding per pair, all floats will be with right precisions\n- **0.6** - included changes from PR kudos to https://github.com/samueltayishere, fixed limit orders, fixed timeouts, fixed is_active order status\n- **0.5.1** - added symbols YFI, BAND, fixed test with limit orders\n- **0.5** - missing symbols MKR, UNI, possible refactoring for simple objects\n- **0.4.5** - fixed missing CELR balances\n- **0.4.4** - fixed missing QTUM, CELR coins\n- **0.4.3** - fixed missing `fees_coin` Coin enum\n- **0.4.2** - fixed supported pairs OMG and MANA\n- **0.4.1** - fixed `cached_property` for python 3.7\n- **0.4.0** - added `OrderForceType` and `OrderExecType`, refactored `Order` responses, splited private and market methods, added missing `Pair` and `Coin`, added `Balance` dataclass, public\nkeys for tests passing\n- **0.3.4** - fixed balances listener, fixed pairs\n- **0.3.3** - fixed orderbook depth\n- **0.3.2** - added orderbook websocket method\n- **0.3.1** - fixed missing DAI pair\n- **0.3** - added websocket support for public endpoints and supports `sign=True` for private endpoints\n- **0.2.1** - fixed order_id in `get_order` func, still preparing for stable release\n- **0.2** - moved to new API v2, except for websockets\n",
    'author': 'Morty Space',
    'author_email': 'morty.space@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import AsyncGenerator, Dict, List

from . import pairs
from .api import ApiProvider
from .structs import (
    Candle,
    MarketTicker,
    MarketTrade,
    OrderBook,
    OrderInBook,
    OrderSide,
    Pair,
    Period,
)


class Exchange:
    """Interface to base exchange methods."""

    def __init__(self, api: ApiProvider = None):
        self.api = api or ApiProvider(auth_required=False)
        self.pairs = {pair.name: pair for pair in pairs.all()}

    async def sync_pairs(self):
        """Use this method to sync pairs if you have issues with missing
        pairs in library side."""
        self.pairs = {pair.name: pair for pair in (await self.get_pairs())}

    async def get_pairs(self) -> List[Pair]:
        """List all available market pairs and store to provide pairs info."""
        data = await self.api.get("public/get-instruments")
        return [
            Pair(
                i["instrument_name"],
                price_precision=i["price_decimals"],
                quantity_precision=i["quantity_decimals"],
            )
            for i in data["instruments"]
        ]

    async def get_ticker(self, pair: Pair) -> MarketTicker:
        """Get ticker in for provided pair."""
        data = await self.api.get(
            "public/get-ticker", {"instrument_name": pair.name}
        )
        return MarketTicker.from_api(pair, data)

    async def get_tickers(self) -> Dict[Pair, MarketTicker]:
        """Get tickers in all available markets."""
        data = await self.api.get("public/get-ticker")
        return {
            self.pairs[ticker["i"]]: MarketTicker.from_api(
                self.pairs[ticker["i"]], ticker
            )
            for ticker in data
            if ticker["i"] in self.pairs
        }

    async def get_price(self, pair: Pair) -> float:
        """Get latest price of pair."""
        return (await self.get_ticker(pair)).trade_price

    async def get_trades(self, pair: Pair) -> List[MarketTrade]:
        """Get last 200 trades in a specified market."""
        data = await self.api.get(
            "public/get-trades", {"instrument_name": pair.name}
        )
        return [MarketTrade.from_api(pair, trade) for trade in reversed(data)]

    async def get_orderbook(self, pair: Pair) -> OrderBook:
        """Get the order book for a particular market, depth always 150."""
        data = await self.api.get(
            "public/get-book", {"instrument_name": pair.name}
        )
        buys = [
            OrderInBook(*order, pair, OrderSide.BUY)
            for order in data[0]["bids"]
        ]
        sells = [
            OrderInBook(*order, pair, OrderSide.SELL)
            for order in reversed(data[0]["asks"])
        ]
        return OrderBook(buys, sells, pair)

    async def get_candles(self, pair: Pair, period: Period) -> List[Candle]:
        data = await self.api.get(
            "public/get-candlestick",
            {"instrument_name": pair.name, "timeframe": period.value},
        )
        return [Candle.from_api(pair, candle) for candle in data]

    async def listen_candles(
        self, period: Period, *pairs: List[Pair]
    ) -> AsyncGenerator[Candle, None]:
        if not isinstance(period, Period):
            raise ValueError(f"Provide Period enum not {period}")

        channels = [f"candlestick.{period}.{pair.name}" for pair in pairs]

        async for data in self.api.listen("market", *channels):
            pair = self.pairs[data["instrument_name"]]
            for candle in data["data"]:
                yield Candle.from_api(pair, candle)

    async def listen_trades(self, *pairs: List[Pair]) -> MarketTrade:
        channels = [f"trade.{pair.name}" for pair in pairs]
        async for data in self.api.listen("market", *channels):
            for trade in data["data"]:
                pair = self.pairs[data["instrument_name"]]
                yield MarketTrade.from_api(pair, trade)

    async def listen_orderbook(self, *pairs: List[Pair]) -> OrderBook:
        channels = [f"book.{pair.name}.150" for pair in pairs]
        async for data in self.api.listen("market", *channels):
            pair = self.pairs[data["instrument_name"]]
            buys = [
                OrderInBook(*order, pair, OrderSide.BUY)
                for order in data["data"][0]["bids"]
            ]
            sells = [
                OrderInBook(*order, pair, OrderSide.SELL)
                for order in reversed(data["data"][0]["asks"])
            ]
            yield OrderBook(buys, sells, pair)
